
 -- This program demonstrates the implementation of several game
 -- NPCs (non-player characters), and how Topaz can be used to
 -- integrate them in with human players in a simplified,
 -- consistent fashion.

 --------------
 -- INCLUDES --
 --------------

include topaz.e
include image.e

without warning

 ---------------
 -- CONSTANTS --
 ---------------

constant HUMAN_CHAR = 1
constant ESC = 27

constant N = 1, E = 2, S = 3, W = 4,
         X = 1, Y = 2,
         N_KEY = '8', S_KEY = '2', E_KEY = '6', W_KEY = '4'

constant EXEC_RATE = 1 / 15 -- everything happens all in 1/15 sec.

constant KEY_RATE  = EXEC_RATE / 2 -- how much of the timeslot to wait
                                   -- for user to press a valid key

constant BARRICADE = 9
constant BARRICADE_COLOR = YELLOW

constant COLORS = {BRIGHT_WHITE,BRIGHT_RED,BRIGHT_BLUE,BRIGHT_GREEN}
constant INIT_POSITIONS = {{3,3}, {23,21}, {10,12}, {18,7}}

constant NUM_THREADS = length (COLORS)

constant ORIENT_SIGNS = {'^', '>', 'v', '<'}
constant CLEAN_DISPLAY_LINE = {254, 0, 254, 0, 254, 0, 254, 0, 254, 0,
                               254, 0, 254, 0, 254, 0, 254, 0, 254, 0,
                               254, 0, 254, 0, 254, 0, 254, 0, 254, 0,
                               254, 0, 254, 0, 254, 0, 254, 0, 254, 0,
                               254, 0, 254, 0, 254, 0, 254, 0, 254, 0}

constant w = WHITE
constant PROGRAM_SCREEN_INFO =
     {{'U',w,'S',w,'E',w,' ',w,
       'N',w,'U',w,'M',w,'E',w,'R',w,'I',w,'C',w},
      {'K',w,'E',w,'Y',w,'P',w,'A',w,'D',w,' ',w,'T',w,'O',w},
      {'M',w,'O',w,'V',w,'E',w,'.',w},
      {' ',w},
      {' ',w},
      {'P',w,'R',w,'E',w,'S',w,'S',w,' ',w,
       '\'',w,'E',w,'S',w,'C',w,'\'',w},
      {'T',w,'O',w,' ',w,'Q',w,'U',w,'I',w,'T',w,'.',w}}

constant INIT_GRID_SPACE =
          {{9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9}} &
   repeat ({9,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,9}, 23) &
          {{9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9}}

 ----------------
 -- LOCAL VARS --
 ----------------

integer not_done, curr_char

sequence grid_space, orientations, positions, thread_ids,
         human_pos, curr_pos, new_pos, new_display,
         new_display_slice, grid_space_slice

 --------------
 -- ROUTINES --
 --------------

function abs (object x)

   return x * (((x > 0) * 2) - 1)

end function -- abs()

 ----------

function invalid (integer key)

   return (key != N_KEY) and (key != S_KEY) and
          (key != E_KEY) and (key != W_KEY) and
          (key != ESC)

end function -- invalid()

 ----------

function get_orientation ()

   curr_char = find (get_thread_id (), thread_ids)

   return orientations[curr_char]

end function -- get_orientation()

 ----------

procedure set_orientation (integer orient)

   curr_char = find (get_thread_id (), thread_ids)

   orientations[curr_char] = orient

end procedure -- set_orientation()

 ----------

function get_char_pos ()

   curr_char = find (get_thread_id (), thread_ids)

   return positions[curr_char]

end function -- get_char_pos()

 ----------

function get_user_pos ()

   return human_pos

end function -- get_user_pos()

 ----------

procedure move_char (integer direction)

   curr_char = find (get_thread_id (), thread_ids)
   curr_pos  = positions[curr_char]

    -- no need to correct the new position, since we know a
    -- border surrounds the area, preventing an illegal index
   if (direction = N) then
      new_pos = curr_pos - {0,1}
   elsif (direction = S) then
      new_pos = curr_pos + {0,1}
   elsif (direction = E) then
      new_pos = curr_pos + {1,0}
   else
      new_pos = curr_pos - {1,0}
   end if

   if (grid_space[new_pos[Y]][new_pos[X]] = 0) then
      grid_space[curr_pos[Y]][curr_pos[X]] = 0
      grid_space[new_pos[Y]][new_pos[X]] = curr_char
      positions[curr_char] = new_pos
      if (curr_char = HUMAN_CHAR) then
         human_pos = new_pos
      end if
   end if

end procedure -- move_char()

 ----------

procedure refresh_screen ()

   integer spot

   new_display = repeat (CLEAN_DISPLAY_LINE, 25)

   for y = 1 to 25 do
   new_display_slice = new_display[y]
   grid_space_slice  = grid_space[y]
   for x = 1 to 25 do

      spot = grid_space_slice[x]
      if (spot = 0) then
          -- do nothing
      elsif (spot = BARRICADE) then
         new_display_slice[x+x]   = BARRICADE_COLOR
      else
         new_display_slice[x+x]   = WHITE + (COLORS[spot] * 16)
         new_display_slice[x+x-1] = ORIENT_SIGNS[orientations[spot]]
      end if

   end for
   new_display[y] = new_display_slice
   end for

   display_text_image ({1,1}, new_display)

end procedure -- refresh_screen()

 ----------

procedure refresh_entire_screen ()

   refresh_screen ()
   display_text_image ({10,28}, PROGRAM_SCREEN_INFO)

end procedure -- refresh_entire_screen ()

 ----------------------------------
 -- PROCS FOR TOPAZ INSTRUCTIONS --
 ----------------------------------

procedure Nothing (sequence params)

    -- this procedure is intentionally left empty

end procedure -- Nothing()

 ----------

procedure Forward (sequence params)

   move_char (get_orientation ())

end procedure -- Forward()

 ----------

procedure TurnClock (sequence params)

   if (get_orientation () = W) then
      set_orientation (N)
   else
      set_orientation (get_orientation () + 1)
   end if

end procedure -- TurnClock()

 ----------

procedure TurnCounter (sequence params)

   if (get_orientation () = N) then
      set_orientation (W)
   else
      set_orientation (get_orientation () - 1)
   end if

end procedure -- TurnCounter()

 ----------

procedure GoN (sequence params)

   set_orientation (N)
   move_char (N)

end procedure -- GoN()

 ----------

procedure GoS (sequence params)

   set_orientation (S)
   move_char (S)

end procedure -- GoS()

 ----------

procedure GoE (sequence params)

   set_orientation (E)
   move_char (E)

end procedure -- GoE()

 ----------

procedure GoW (sequence params)

   set_orientation (W)
   move_char (W)

end procedure -- GoW()

 ----------

procedure Wander (sequence params)

   integer r
   r = rand (15)
   if (r <= 3) then
      Nothing (params)
   elsif (r <= 11) then
      Forward (params)
   elsif (r = 12) then
      GoN (params)
   elsif (r = 13) then
      GoE (params)
   elsif (r = 14) then
      GoS (params)
   else
      GoW (params)
   end if

end procedure -- Wander()

 ----------

procedure Follow (sequence params)

   sequence xy1, xy2
   integer  diffx, diffy, abs_diffx, abs_diffy

   xy1 = get_char_pos ()
   xy2 = get_user_pos ()

   diffx = xy1[X] - xy2[X]
   diffy = xy1[Y] - xy2[Y]
   abs_diffx = abs (diffx)
   abs_diffy = abs (diffy)

   if (abs_diffx + abs_diffy > 3) then
      if (abs_diffy > abs_diffx) then
         if (diffy < 0) then
            set_orientation (S)
            move_char (S)
         else
            set_orientation (N)
            move_char (N)
         end if
      else
         if (diffx < 0) then
            set_orientation (E)
            move_char (E)
         else
            set_orientation (W)
            move_char (W)
         end if
      end if
   end if

end procedure -- Follow()

 ----------

procedure GetUserCmd (sequence params)

   integer k
   atom    begin_key_wait

    -- if we don't wait a while for
    -- the keypress, we might miss it.
   k = get_key ()
   begin_key_wait = time ()
   while (invalid (k)) and (begin_key_wait + KEY_RATE > time ()) do
      k = get_key ()
   end while

   if (not (invalid (k))) then
      if (k = N_KEY) then
         move_char (N)
      elsif (k = S_KEY) then
         move_char (S)
      elsif (k = E_KEY) then
         move_char (E)
      elsif (k = W_KEY) then
         move_char (W)
      else
         not_done = 0 -- it must be ESC, so quit
      end if
   end if

end procedure -- GetUserCmd()


 ----------------------
 -- NEW INSTRUCTIONS --
 ----------------------

 -- user-defined instructions
constant tFORWARD      = add_instruction (routine_id ("Forward"))
constant tTURN_CLOCK   = add_instruction (routine_id ("TurnClock"))
constant tTURN_COUNTER = add_instruction (routine_id ("TurnCounter"))
constant tWANDER       = add_instruction (routine_id ("Wander"))
constant tNOTHING      = add_instruction (routine_id ("Nothing"))
constant tFOLLOW       = add_instruction (routine_id ("Follow"))
constant tGO_N         = add_instruction (routine_id ("GoN"))
constant tGO_S         = add_instruction (routine_id ("GoS"))
constant tGO_E         = add_instruction (routine_id ("GoE"))
constant tGO_W         = add_instruction (routine_id ("GoW"))
constant tGET_USER_CMD = add_instruction (routine_id ("GetUserCmd"))

 -- The scripts for the different types of characters,
 -- including human-controlled. Note that all of the
 -- scripts are the same size. Having them be not the same
 -- size, but close, is fine if all are large scripts. But
 -- a 2-instruction script, where the last instruction loops
 -- to the top, has an enormous advantage over, say, a
 -- 5-instruction script that ends with a loop back to the
 -- first instruction.

constant WANDER_SCRIPT = -- script for comp. block that just wanders
                         {{tWANDER},
                          {tWANDER},
                          {tWANDER},
                          {tWANDER},
                          {tWANDER},
                          {tWANDER},
                          {tINCIP, -6}}

constant SEEK_SCRIPT   = -- for comp. block that follows you, at dist.
                         {{tFOLLOW},
                          {tNOTHING},
                          {tFOLLOW},
                          {tNOTHING},
                          {tFOLLOW},
                          {tNOTHING},
                          {tINCIP, -6}}

constant CIRCLE_SCRIPT = -- for comp. block that goes in circles
                         {{tTURN_CLOCK},
                          {tFORWARD},
                          {tFORWARD},
                          {tFORWARD},
                          {tFORWARD},
                          {tFORWARD},
                          {tINCIP, -6}}

constant HUMAN_SCRIPT  = -- for human-controlled block
                         {{tGET_USER_CMD},
                          {tGET_USER_CMD},
                          {tGET_USER_CMD},
                          {tGET_USER_CMD},
                          {tGET_USER_CMD},
                          {tGET_USER_CMD},
                          {tINCIP, -6}}

 ----------
 -- MAIN --
 ----------

 -- This program illustrates the implementation of background-active
 -- NPCs. Note that the computer-controlled characters all move about
 -- regardless of your actions, and that they do so at a constant
 -- speed, whether you're moving or not.

procedure main ()

   atom start
   object junk

   junk = graphics_mode (1)

   grid_space   = INIT_GRID_SPACE
   positions    = INIT_POSITIONS
   orientations = repeat (N, 4)
   human_pos    = positions[HUMAN_CHAR]

   for i = 1 to NUM_THREADS do
      grid_space[positions[i][Y]][positions[i][X]] = i
   end for

    -- Actually create all of the threads (priorities
    -- set for 1, since we want them to be at the max.)
    -- Try adjusting the priorities for diff. effects!
   thread_ids = {new_thread (0, 1, HUMAN_SCRIPT),
                 new_thread (0, 1, SEEK_SCRIPT),
                 new_thread (0, 1, CIRCLE_SCRIPT),
                 new_thread (0, 1, WANDER_SCRIPT)}

   refresh_entire_screen ()

    -- loop through all threads at a constant rate
   tick_rate (100) -- allow for more accurate time evaluation
   not_done = 1
   while (not_done) do
      start = time ()
      for i = 1 to NUM_THREADS do
         topaz_step ()
      end for
      refresh_screen ()
       -- this delay makes sure that the computer chars don't
       -- go faster just because you're not pressing a key
      while ((start + EXEC_RATE) > time ()) do
      end while
   end while
   tick_rate (0) -- bring back to normal

    -- we're done
   for i = 1 to NUM_THREADS do
      kill_thread (thread_ids[i])
   end for

   junk = graphics_mode (3)

end procedure -- main ()

 ----------
 -- EXEC --
 ----------

main ()

