
 ---------------------------------------------------------------------
 --                                                                 --
 --                        Topaz v. 2.1                             --
 --                                                                 --
 --            Imbedded Stack-Based Scripting Engine                --
 --                    (a "virtual machine")                        --
 --                                                                 --
 --       Created July, Dec 1999; March 2002 by Rod Jackson.        --
 --                     All rights reserved.                        --
 --                                                                 --
 -- This source code is made available free of charge for all       --
 -- public, private, and commercial use. It has been designed to    --
 -- run under version 2.0+ of the Euphoria interpreter by Rapid     --
 -- Development Software. Only authorized modifications or upgrades --
 -- should be distributed unless said modifications or upgrades are --
 -- clearly identified as "non-official" or "unofficial".           --
 --                                                                 --
 -- For more information on this library, please refer to the       --
 -- documentation files, or contact the author at:                  --
 --                                                                 --
 --    rodjackson@bigfoot.com                                       --
 --                                                                 --
 ---------------------------------------------------------------------

 ---------------------
 -- local constants --
 ---------------------

constant MAX_EU_INT = 1073741823 -- the max. legal size of an integer
constant TOPAZ_ABORT_CODE = 'T'  -- this is 84 decimal, 54 hex

 -----------------
 -- local types --
 -----------------

 -- a legal virtual dataspace address
type vaddress (object x)
   if (integer (x)) then
      if (x > 0) then
         return 1 -- TRUE
      end if
   end if
   return 0 -- FALSE
end type

 ---------------------
 -- local variables --
 ---------------------

sequence InstructionSpaceTable,   -- the instruction space per thread
         DataSpaceTable,          -- the data space of each thread
         StackTable,              -- the stack of each thread
         StackSizeTable,          -- the size of each thread's stack
         ThreadIDTable,           -- a unique ID per thread
         InstructionPointerTable, -- next instr. to exec. per thread
         InstructionRoutineTable, -- routine ID per instruction
         IsSuspendedTable,        -- if each thread is suspended
         TruePriorityTable,       -- each thread's original priority
         CalcPriorityTable        -- each thread's calc. priority
                                  -- (used to make them relative)

integer  MaxInstruction,      -- the highest value of all instructions
         ThreadPointer,       -- indicates the current thread
         TP_IsTrapped,        -- if the thread pointer is trapped
         TP_TrapsAllowed,     -- if thread pointer trapping is allowed
         NumThreads,          -- the number of existing threads
         NextThreadID,        -- the value to assign a new thread's ID
         CurrentIP,           -- the IP of the current thread
         IPWasChanged,        -- if the curr. instr. for the thread
                              --   altered the instruction pointer
         CurrThreadIsFinished,-- if the current thread has ended
         CurrPriorityStep,    -- determines priorities to get exec.
                              -- this round
         HighestCalcPriority  -- highest calculated priority

atom     L -- used to hold sequence lengths for local functions

 --------------------
 -- local routines --
 --------------------

 --
 -- Min: Returns the minimum value in a sequence.
 --
atom the_min, min_curr -- declared locally for speed
function Min (sequence s)

   the_min = s[1]
   for i = 2 to length (s) do
      min_curr = s[i]
      if (min_curr < the_min) then
         the_min = min_curr
      end if
   end for

   return the_min

end function -- Min()


 --
 -- Max: Returns the maximum value in a sequence.
 --
atom the_max, max_curr -- declared locally for speed
function Max (sequence s)

   the_max = s[1]
   for i = 2 to length (s) do
      max_curr = s[i]
      if (max_curr > the_max) then
         the_max = max_curr
      end if
   end for

   return the_max

end function -- Max()


 --
 -- Remove: Returns the given sequence, with the specified element
 --    removed.
 --
function Remove (integer pos, sequence s)

   return s[1..pos-1] & s[pos+1..length(s)]

end function -- Remove()


 --
 -- set_element: Given a sequence, the index list to descend to, and
 --    a replacement object, it will replace the specified element of
 --    said sequence with said object.
 --
function set_element (sequence s, sequence indices, object obj)

   L = length (indices)

   if (L = 1) then
      s[indices[1]] = obj
      return s
   else
      s[indices[1]] = set_element (s[indices[1]], indices[2..L], obj)
      return s
   end if

end function -- set_element()


 --
 -- IncrementThreadID: Increments the value of NextThreadID so that
 --    any new threads will be given a unique ID.
 --
procedure IncrementThreadID ()

   integer AttemptedNext

   if (NextThreadID = MAX_EU_INT) then
      AttemptedNext = 0
   else
      AttemptedNext = NextThreadID + 1
   end if

    -- make sure the ID isn't already taken (highly unlikely, but...)
   while (find (AttemptedNext, ThreadIDTable)) do
      if (AttemptedNext = MAX_EU_INT) then
         AttemptedNext = 0
      else
         AttemptedNext = AttemptedNext + 1
      end if
   end while

   NextThreadID = AttemptedNext

end procedure -- IncrementThreadID()


 --
 -- IncrementThreadPointer: Increments the thread pointer so that the
 --    next thread will be executed. Cycles back to the first after
 --    running through all existing threads.
 --
atom priority_check -- declared locally for speed
procedure IncrementThreadPointer ()

   ThreadPointer = ThreadPointer + 1

   if (ThreadPointer > NumThreads) then
      ThreadPointer = 1
      CurrPriorityStep = CurrPriorityStep - 1
      if (CurrPriorityStep < 1) then
         CurrPriorityStep = HighestCalcPriority
      end if
   end if

    -- make sure we actually have a thread to worry about first!
   if (NumThreads > 0) then

      priority_check = CurrPriorityStep /
                       CalcPriorityTable[ThreadPointer]

      while (priority_check != floor (priority_check)) do
         ThreadPointer = ThreadPointer + 1
         if (ThreadPointer > NumThreads) then
            ThreadPointer = 1
            CurrPriorityStep = CurrPriorityStep - 1
            if (CurrPriorityStep < 1) then
               CurrPriorityStep = HighestCalcPriority
            end if
         end if
         priority_check = CurrPriorityStep /
                          CalcPriorityTable[ThreadPointer]
      end while

   end if -- check for one or more threads

end procedure -- IncrementThreadPointer()


 --
 -- InitTopaz: Initializes the internal variables and tables of the
 --    Topaz engine.
 --
procedure InitTopaz ()

    -- initialize the thread-local data tables
   InstructionSpaceTable   = {}
   DataSpaceTable          = {}
   StackTable              = {}
   StackSizeTable          = {}
   ThreadIDTable           = {}
   InstructionPointerTable = {}
   IsSuspendedTable        = {}
   TruePriorityTable       = {}
   CalcPriorityTable       = {}

    -- initialize the instruction code table
   InstructionRoutineTable = {}

    -- initialize the internal Topaz variables
   MaxInstruction       = 0
   ThreadPointer        = 1
   TP_IsTrapped         = 0 -- FALSE
   TP_TrapsAllowed      = 1 -- TRUE
   NumThreads           = 0
   NextThreadID         = 0
   CurrentIP            = 0
   IPWasChanged         = 0 -- FALSE
   CurrThreadIsFinished = 0 -- FALSE
   CurrPriorityStep     = 1

end procedure -- InitTopaz()


 --
 -- TopazError: Displays a Topaz-specific error message in a
 --    standardized format, then aborts with a Topaz-specific error
 --    code.
 --
procedure TopazError (integer  LorPorF,     -- if func., proc. or lib.
                      sequence RoutineName, -- name of routine error
                                            --   was found in
                      sequence ErrMessage)  -- the specific error
                                            --   message

   sequence EntireMessage

   EntireMessage = "\nTopaz error in "

   if (LorPorF = 'l') or (LorPorF = 'L') then
      EntireMessage = EntireMessage & "library -\n   " &
                      ErrMessage & "\n"
   else
      if (LorPorF = 'p') or (LorPorF = 'P') then
         EntireMessage = EntireMessage & "procedure"
      else
         EntireMessage = EntireMessage & "function"
      end if

      EntireMessage = EntireMessage & ": " & RoutineName & " -\n   " &
                      ErrMessage & "\n"
   end if

   puts (1, EntireMessage)

   abort (TOPAZ_ABORT_CODE)

end procedure -- TopazError()


 --
 -- TopazExec: Executes the given instruction by calling the
 --    appropriate routine and passing all parameters.
 --
object TheInstr -- declared locally for speed
procedure TopazExec (sequence FullInstr)

   TheInstr = FullInstr[1]

   if (not (integer (TheInstr))) then
      TopazError ('p', "TopazExec", "Unknown instruction encountered")
   end if

   call_proc (InstructionRoutineTable[TheInstr],
              {FullInstr[2..length(FullInstr)]})

end procedure -- TopazExec()


 --
 -- RecalcPriorities: Adjusts the working priorities of the threads
 --    so that they all maintain the same rate of execution relative
 --    to each other.
 --
integer low -- declared locally for speed
procedure RecalcPriorities ()

   if (length (TruePriorityTable) < 2) then
      if (length (TruePriorityTable) = 1) then
         CalcPriorityTable = {1}
      else
         CalcPriorityTable = {}
      end if
      HighestCalcPriority = 1
      CurrPriorityStep = 1
   else
      low = Min (TruePriorityTable)
      if (low = 1) then
         CalcPriorityTable = floor (TruePriorityTable)
      else
         CalcPriorityTable = floor ((TruePriorityTable / low) + 0.5)
      end if
      HighestCalcPriority = Max (CalcPriorityTable)
      CurrPriorityStep = HighestCalcPriority
   end if

end procedure -- RecalcPriorities()


 ------------------------------
 -- primary library routines --
 ------------------------------

 --
 -- add_instruction: Adds a new instruction to the Topaz engine by
 --    registering the routine ID of the code that will be executed
 --    when the instruction is encountered. After being added, the
 --    instruction can immediately be utilized.
 --
global function add_instruction (integer RoutineID)

   MaxInstruction = MaxInstruction + 1
   InstructionRoutineTable = InstructionRoutineTable & RoutineID
   return MaxInstruction

end function -- add_instruction()


 --
 -- allow_tp_traps: Tells the Topaz engine to honor all subsequent
 --    requests to trap the thread pointer.
 --
global procedure allow_tp_traps ()

   TP_TrapsAllowed = 1 -- TRUE

end procedure -- allow_tp_traps()


 --
 -- disallow_tp_traps: Tells the Topaz engine to untrap the thread
 --    pointer if trapped, and to ignore all subsequent requests
 --    to trap it.
 --
global procedure disallow_tp_traps ()

   TP_TrapsAllowed = 0 -- FALSE
   TP_IsTrapped    = 0 -- FALSE

end procedure -- disallow_tp_traps()


 --
 -- is_tp_trapped: Returns 1 (TRUE) if the thread pointer is trapped,
 --    otherwise returning 0 (FALSE).
 --
global function is_tp_trapped ()

   return TP_IsTrapped

end function -- is_tp_trapped()


 --
 -- is_stack_empty: Returns 1 (TRUE) if the current thread's stack is
 --    empty, otherwise returning 0 (FALSE).
 --
global function is_stack_empty ()

   return (length (StackTable[ThreadPointer]) = 0)

end function -- is_stack_empty()


 --
 -- get_vaddress: Returns the object at the specified virtual address
 --    of the data space of the current thread.
 --
global function get_vaddress (vaddress Addr)

   return DataSpaceTable[ThreadPointer][Addr]

end function -- get_vaddress()


 --
 -- put_vaddress: Puts the given object into the data space of the
 --    current thread, at the specified virtual address.
 --
global procedure put_vaddress (object Obj, vaddress Addr)

   DataSpaceTable[ThreadPointer][Addr] = Obj

end procedure -- put_vaddress()


 --
 -- pop_stack: Pops the top element of the stack of the current
 --    thread, shrinking the stack. The popped element is returned.
 --
global function pop_stack ()

   object   temp
   integer  stacksize

   stacksize = StackSizeTable[ThreadPointer]

   temp = StackTable[ThreadPointer][stacksize]

   stacksize = stacksize - 1

   StackTable[ThreadPointer] =
         StackTable[ThreadPointer][1..stacksize]

   StackSizeTable[ThreadPointer] = stacksize

   return temp

end function -- pop_stack()


 --
 -- push_stack: Pushes the given object onto the stack of the current
 --    thread.
 --
global procedure push_stack (object Obj)

   StackTable[ThreadPointer] = append (StackTable[ThreadPointer], Obj)

   StackSizeTable[ThreadPointer] = StackSizeTable[ThreadPointer] + 1

end procedure -- push_stack()


 --
 -- new_thread: Creates a new thread, giving it its own data space,
 --    instruction space, and stack. The code to be placed into the
 --    instruction space must be provided, as must the size of the
 --    data space (in number of elements) and the priority (1+).
 --
global function new_thread (integer  DataSpaceSize,
                            integer  Priority,
                            sequence InstructionCode)

   integer ThreadID

   if (Priority < 1) then
      TopazError ('f', "new_thread", "Illegal priority; too low")
   end if

   ThreadID = NextThreadID
   ThreadIDTable = ThreadIDTable & ThreadID
   IncrementThreadID ()

   InstructionSpaceTable = append
                             (InstructionSpaceTable, InstructionCode)
   DataSpaceTable = append (DataSpaceTable, repeat (0, DataSpaceSize))
   InstructionPointerTable = InstructionPointerTable & 1
   StackTable = append (StackTable, {})
   StackSizeTable = StackSizeTable & 0
   IsSuspendedTable = IsSuspendedTable & 0 -- FALSE
   TruePriorityTable = TruePriorityTable & Priority

   NumThreads = NumThreads + 1

   RecalcPriorities ()

   return ThreadID

end function -- new_thread()


 --
 -- kill_thread: kills the specified thread, removing its stack,
 --    instruction space and data space.
 --
global procedure kill_thread (integer ThreadID)

   integer thread_pos

    -- find out at what position the specified thread is found
   thread_pos = find (ThreadID, ThreadIDTable)

    -- make sure the thread ID given is legal
   if (thread_pos = 0) then
      TopazError ("f", "kill_thread", "Unknown thread ID given")
   end if

    -- remove the local data for the thread from the tables
   InstructionSpaceTable = Remove (thread_pos, InstructionSpaceTable)
   InstructionPointerTable = Remove (thread_pos,
                                     InstructionPointerTable)
   StackTable = Remove (thread_pos, StackTable)
   StackSizeTable = Remove (thread_pos, StackSizeTable)
   DataSpaceTable = Remove (thread_pos, DataSpaceTable)
   ThreadIDTable = Remove (thread_pos, ThreadIDTable)
   IsSuspendedTable = Remove (thread_pos, IsSuspendedTable)
   TruePriorityTable = Remove (thread_pos, TruePriorityTable)
    -- no need to remove it from CalcPriorityTable; this will
    -- be recalculated from TruePriorityTable during RecalcPriorities

    -- adjust & untrap the thread pointer if necessary
   if (thread_pos < ThreadPointer) then
      ThreadPointer = ThreadPointer - 1
   elsif (thread_pos = ThreadPointer) then
      TP_IsTrapped = 0 -- FALSE
   end if

    -- note that the number of threads is reduced
   NumThreads = NumThreads - 1

    -- recalculate thread priorities
   RecalcPriorities ()

    -- make sure we're not past the threads
   if (ThreadPointer > NumThreads) then
      IncrementThreadPointer () -- takes care of cases > NumThreads
   end if

   CurrThreadIsFinished = 0 -- FALSE

end procedure -- kill_thread()


 --
 -- get_thread_id: Returns the unique ID of the current thread.
 --
global function get_thread_id ()

   return ThreadIDTable[ThreadPointer]

end function -- get_thread_id()


 --
 -- get_num_threads: Returns the total number of existant threads.
 --
global function get_num_threads ()

   return NumThreads

end function -- get_num_threads()


 --
 -- inc_ip: Increments the current instruction pointer by the
 --    specified amount, and notes that the instruction pointer was
 --    modified.
 --
global procedure inc_ip (integer increment)

   CurrentIP = CurrentIP + increment

   IPWasChanged = 1 -- TRUE

end procedure -- inc_ip()


 --
 -- topaz_step: Executes the next instruction of the current thread,
 --    then increments the thread pointer so that next time the next
 --    thread will be executed. If the thread pointer is trapped, it
 --    will not be incremented.
 --
global procedure topaz_step ()

   object FullInstr

    -- only execute if thread pointer is valid (if not, most likely
    -- there are no existing threads, which we'll handle), and...
   if (ThreadPointer <= NumThreads) then

       -- ...if the thread is not suspended
      if (IsSuspendedTable[ThreadPointer] = 0) then

         CurrentIP = InstructionPointerTable[ThreadPointer]

         if (CurrentIP < 0) then
             -- this means the last command was an 'exec', so we'll
             -- get the code from the stack
            CurrentIP = 0 - CurrentIP
            FullInstr = pop_stack ()
         else
             -- just execute the next instruction in instruction space
            FullInstr =InstructionSpaceTable[ThreadPointer][CurrentIP]
         end if

          -- make sure the actual instruction is also valid
         if (atom (FullInstr)) then
            TopazError ('p', "topaz_step", "Attempt to execute an " &
                        "improperly formed instruction")
         end if

         IPWasChanged = 0 -- FALSE
         TopazExec (FullInstr)

      end if -- end of check for suspended thread

      if (CurrThreadIsFinished) then
          -- we hit a 'tEND' instruction; kill this thread
          -- (TP is automatically incremented & untrapped)
         kill_thread (get_thread_id())
      else
          -- no 'tEND' was hit, so continue as normal
          -- increment the IP (unless already modified)
         if (not IPWasChanged) then
            if (CurrentIP > 0) then -- otherwise, we encountered tIF
               CurrentIP = CurrentIP + 1
            end if
         end if
          -- save the new IP
         InstructionPointerTable[ThreadPointer] = CurrentIP
          -- increment the TP (unless trapped)
         if (not TP_IsTrapped) then
            IncrementThreadPointer ()
         end if
      end if

   else

       -- the thread pointer is invalid (likely no threads);
       -- the increment routine will fix it if necessary
      IncrementThreadPointer ()

   end if -- end of check for valid thread pointer

end procedure -- topaz_step()

 --------------------------------------------------
 -- local routines for the built-in instructions --
 --------------------------------------------------

without warning

atom   m, n
object Math1, Math2, Obj1, Obj2

 --
 -- tADD: Pops the top two elements of the stack, adds them together,
 --    and pushes the result back onto the stack. No parameters.
 --
procedure add (sequence Params)

   Math1 = pop_stack ()
   Math2 = pop_stack ()

   push_stack (Math1 + Math2)

end procedure -- add()


 --
 -- tSUB: Pops the top two elements of the stack, subtracts the first
 --    from the second, and pushes the result back onto the stack. No
 --    params.
 --
procedure sub (sequence Params)

   Math1 = pop_stack ()
   Math2 = pop_stack ()

   push_stack (Math1 - Math2)

end procedure -- sub()


 --
 -- tMUL: Pops the top two elements of the stack, multiplies them
 --    together, and pushes the result back onto the stack. No params.
 --
procedure mul (sequence Params)

   Math1 = pop_stack ()
   Math2 = pop_stack ()

   push_stack (Math1 * Math2)

end procedure -- mul()


 --
 -- tDIV: Pops the top two elements of the stack, divides the first by
 --    the second, and pushes the result back onto the stack. No
 --    params.
 --
procedure div (sequence Params)

   Math1 = pop_stack ()
   Math2 = pop_stack ()

   push_stack (Math1 / Math2)

end procedure -- div()


 --
 -- tFLOOR: Pops the top element of the stack, performs a Euphorian
 --    'floor' on it, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_floor (sequence Params)

   push_stack (floor (pop_stack ()))

end procedure -- topaz_floor()


 --
 -- tCAT: Pops the top two elements of the stack, concatenates them
 --    (applies a Euphorian '&'), and pushes the result back onto the
 --    stack. No params.
 --
procedure seq_concat (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 & Obj2)

end procedure -- seq_concat()


 --
 -- tEQ: Pops the top two elements of the stack, performs a
 --    Euphorian '=' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure equal_sign (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 = Obj2)

end procedure -- equal_sign()


 --
 -- tLESS: Pops the top two elements of the stack, performs a
 --    Euphorian '<' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure less_than (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 < Obj2)

end procedure -- less_than()


 --
 -- tGRTR: Pops the top two elements of the stack, performs a
 --    Euphorian '>' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure greater_than (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 > Obj2)

end procedure -- greater_than()


 --
 -- tNEQ: Pops the top two elements of the stack, performs a
 --    Euphorian '!=' on them, and pushes the result back onto the
 --    stack. No params.
 --
procedure not_equal (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 != Obj2)

end procedure -- not_equal()


 --
 -- tAND: Pops the top two elements of the stack, applies a Euphorian
 --    'and' on them, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_and (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 and Obj2)

end procedure -- topaz_and()


 --
 -- tOR: Pops the top two elements of the stack, applies a Euphorian
 --    'or' on them, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_or (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 or Obj2)

end procedure -- topaz_or()


 --
 -- tXOR: Pops the top two elements of the stack, applies a Euphorian
 --    'xor' on them, and pushes the result back onto the stack. No
 --    params.
 --
procedure topaz_xor (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (Obj1 xor Obj2)

end procedure -- topaz_xor()


 --
 -- tNOT: Pops the top element of the stack, applies a Euphorian 'not'
 --    to it, and pushes the result back onto the stack. No params.
 --
procedure topaz_not (sequence Params)

   Obj1 = pop_stack ()

   push_stack (not (Obj1))

end procedure -- topaz_not()


 --
 -- tCOMP: Pops the top two elements of the stack, compares them, and
 --    pushes the result (-1, 0, or 1) back onto the stack. No params.
 --
procedure topaz_compare (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (compare (Obj1, Obj2))

end procedure -- topaz_compare()


 --
 -- tEQUAL: Pops the top two elements of the stack, compares them,
 --    and pushes 1 (TRUE) back onto the stack if they're identical
 --    objects, otherwise pushing back a 0 (FALSE). No params.
 --
procedure topaz_equal (sequence Params)

   Obj1 = pop_stack ()
   Obj2 = pop_stack ()

   push_stack (compare (Obj1, Obj2) = 0)

end procedure -- topaz_equal()


 --
 -- tPUSHA: Takes the data located at the specified virtual address,
 --    and pushes it onto the stack. If there are multiple addresses,
 --    this is repeated for each. One or more params.
 --
procedure push_vaddr (sequence Params)

   for i = 1 to length (Params) do
      push_stack (get_vaddress(Params[i]))
   end for

end procedure -- push_vaddr()


 --
 -- tPUSHD: Takes the parameter, and pushes it onto the stack. If
 --    there is more than one, this is repeated for each. One or more
 --    params.
 --
procedure push_data (sequence Params)

   for i = 1 to length (Params) do
      push_stack (Params[i])
   end for

end procedure -- push_data()


 --
 -- tPOP: Pops the top element from the stack and places it in the
 --    specified virtual address. If there are multiple addresses,
 --    the process is repeated for each. If a virtual address is 0,
 --    the top element of the stack is discarded. One or more params.
 --
procedure topaz_pop (sequence Params)

   object  Obj
   integer VAddr

   for i = 1 to length (Params) do
      Obj = pop_stack ()
      VAddr = Params[i]
      if (VAddr != 0) then
         put_vaddress (Obj, VAddr)
      end if
   end for

end procedure -- topaz_pop()


 --
 -- tCOPY: Creates a duplicate of the top stack element, and pushes
 --    it onto the stack. No params.
 --
procedure topaz_copy (sequence Params)

   Obj1 = pop_stack ()

   push_stack (Obj1)
   push_stack (Obj1)

end procedure -- topaz_copy()


 --
 -- tIF: Pops the top element of the stack. Then executes its
 --    parameter (which is a single instruction, properly wrapped with
 --    its own parameters) if that element is not zero. One param.
 --
procedure if_not_0 (sequence Params)

   Obj1 = pop_stack ()

   if (atom (Obj1)) then
      if (Obj1 != 0) then
         push_stack (Params[1])
         CurrentIP = 0 - CurrentIP
      end if
   end if

end procedure -- if_not_0()


 --
 -- tTRAP: Traps the thread pointer, causing the Topaz engine to
 --    dedicate all execution steps to the current thread (until it
 --    untraps the pointer). No params.
 --
procedure trap_tp (sequence Params)

   if (TP_TrapsAllowed) then
      TP_IsTrapped = 1 --TRUE
   end if

end procedure -- trap_tp()


 --
 -- tUTRAP: Untraps the thread pointer, allowing other threads to
 --    run concurrently. No params.
 --
procedure untrap_tp (sequence Params)

   TP_IsTrapped = 0 -- FALSE

end procedure -- untrap_tp()


 --
 -- tSUSP: Pops the top element of the stack, then suspends the
 --    thread with matching ID. If the current thread suspends
 --    itself, then the thread pointer is untrapped. No params.
 --
procedure suspend_thread (sequence Params)

   object  ThreadID
   integer TargetThread

   ThreadID = pop_stack ()

   TargetThread = find (ThreadID, ThreadIDTable)

   if (TargetThread = 0) then
      TopazError ('p', "suspend_thread", "Unknown thread ID given")
   end if

   IsSuspendedTable[TargetThread] = 1 -- TRUE

   if (ThreadID = get_thread_id ()) then
      untrap_tp (Params)
   end if

end procedure -- suspend_thread()


 --
 -- tRESUM: Pops the top element of the stack, and identifies the
 --    thread with a matching ID. Then sets the specified thread's
 --    flag so that it is no longer suspended. NOTE: A thread is
 --    incapable of resuming itself.
 --
procedure resume_thread (sequence Params)

   object  ThreadID
   integer TargetThread

   ThreadID = pop_stack ()

   TargetThread = find (ThreadID, ThreadIDTable)

   if (TargetThread = 0) then
      TopazError ('p', "resume_thread", "Unknown thread ID given")
   end if

   IsSuspendedTable[TargetThread] = 0 -- FALSE

end procedure -- resume_thread()


 --
 -- tGETID: Pushes the thread's unique ID onto the stack. No params.
 --
procedure topaz_get_thread_id (sequence Params)

   push_stack (get_thread_id ())

end procedure -- topaz_get_thread_id()


 --
 -- tINCIP: Increments the instruction pointer for the thread by the
 --    amount of that parameter; use negative values to jump/branch
 --    backwards. If the parameter is 0, pops the top element of the
 --    stack and uses that value instead. One param.
 --
procedure topaz_inc_ip (sequence Params)

   object Value

   Value = Params[1]

   if (Value = 0) then
      inc_ip (pop_stack ())
   else
      inc_ip (Value)
   end if

end procedure -- topaz_inc_ip()


 --
 -- tDISP: Pops the top element of the stack and displays it. No
 --    parameters.
 --
procedure topaz_display (sequence Params)

   ? pop_stack ()

end procedure -- topaz_display()


 --
 -- tNOP: This instruction intentionally does nothing. No params.
 --
procedure topaz_nop (sequence Params)

    -- this is meant to be a completely empty procedure

end procedure -- topaz_nop()


 --
 -- tEND: Ends the current thread and causes it to be killed. The
 --    thread pointer is also untrapped. No params.
 --
procedure topaz_end (sequence Params)

   CurrThreadIsFinished = 1 -- TRUE

   untrap_tp (Params)

end procedure -- topaz_end()


 --
 -- tGINDX: Pops a sequence (assumedly) from the top of the stack,
 --    pops the index value(s), and lastly pushes the element at
 --    the specified index. One parameter, which is the number of
 --    indexes to pop. For example, a parameter of 1, with a 4 still
 --    on the stack (after popping the sequence), means to access [4]
 --    while a parameter of 3, with a 2, 9, and then 5 on the stack,
 --    means to access [2][9][5].
 --
procedure get_index (sequence Params)

   object  s
   atom    indx

   s = pop_stack()

   for i = 1 to Params[1] do
      indx = pop_stack()
      s = s[indx]
   end for

   push_stack (s)

end procedure -- get_index()


 --
 -- tSINDX: Pops a sequence (assumedly) from the top of the stack,
 --    pops the index value(s), pops the object to replace with,
 --    sets the specified element to be the replacing object, and
 --    lastly pushes the modified sequence back onto the stack.
 --    One parameter, which is the number of indexes to pop. For
 --    example, a parameter of 1, with a 4 still on the stack (after
 --    popping the sequence), and a {{}} afterward, sets element [4]
 --    to {{}}, while a parameter of 3, with a 2, 9, 5 and then 'q'
 --    on the stack, sets element [2][9][5] to 'q'.
 --
procedure set_index (sequence Params)

   sequence s, index_list

   s = pop_stack()

   index_list = {}
   for i = 1 to Params[1] do
      index_list = append (index_list, pop_stack())
   end for

   Obj1 = pop_stack ()

   s = set_element (s, index_list, Obj1)

   push_stack (s)

end procedure -- set_index()


 --
 -- tGSLIC: Pops a sequence from the top of the stack, pops the slice
 --    boundaries, then pushes the slice of the sequence back onto
 --    the stack. No parameters.
 --
procedure get_slice (sequence Params)

   sequence s

   s = pop_stack()

   m = pop_stack ()
   n = pop_stack ()

   push_stack (s[m..n])

end procedure -- get_slice()


 --
 -- tSSLIC: Pops a sequence from the top of the stack, pops the slice
 --    boundaries, then pops the replacing slice. Replaces the
 --    specified portion of the sequence with the replacing slice,
 --    then pushes the modified sequence back onto the stack.
 --    No parameters.
 --
procedure set_slice (sequence Params)

   sequence s, new_slice

   s = pop_stack()

   m = pop_stack ()
   n = pop_stack ()

   new_slice = pop_stack()

   s[m..n] = new_slice

   push_stack (s)

end procedure -- set_slice()


with warning

 ----------------------
 -- initialize Topaz --
 ----------------------

 -- basic initialization

InitTopaz ()

 -- load the "built-in" instructions & assign to constants
global constant tADD   = add_instruction (routine_id ("add"))
global constant tSUB   = add_instruction (routine_id ("sub"))
global constant tMUL   = add_instruction (routine_id ("mul"))
global constant tDIV   = add_instruction (routine_id ("div"))
global constant tFLOOR = add_instruction (routine_id ("topaz_floor"))
global constant tAND   = add_instruction (routine_id ("topaz_and"))
global constant tCAT   = add_instruction (routine_id ("seq_concat"))
global constant tOR    = add_instruction (routine_id ("topaz_or"))
global constant tXOR   = add_instruction (routine_id ("topaz_xor"))
global constant tNOT   = add_instruction (routine_id ("topaz_not"))
global constant tCOMP  = add_instruction (routine_id ("topaz_compare"))
global constant tEQ    = add_instruction (routine_id ("equal_sign"))
global constant tLESS  = add_instruction (routine_id ("less_than"))
global constant tGRTR  = add_instruction (routine_id ("greater_than"))
global constant tNEQ   = add_instruction (routine_id ("not_equal"))
global constant tPUSHA = add_instruction (routine_id ("push_vaddr"))
global constant tPUSHD = add_instruction (routine_id ("push_data"))
global constant tPOP   = add_instruction (routine_id ("topaz_pop"))
global constant tCOPY  = add_instruction (routine_id ("topaz_copy"))
global constant tIF    = add_instruction (routine_id ("if_not_0"))
global constant tTRAP  = add_instruction (routine_id ("trap_tp"))
global constant tUTRAP = add_instruction (routine_id ("untrap_tp"))
global constant tSUSP  = add_instruction (routine_id ("suspend_thread"))
global constant tRESUM = add_instruction (routine_id ("resume_thread"))
global constant tGETID = add_instruction (routine_id ("topaz_get_thread_id"))
global constant tINCIP = add_instruction (routine_id ("topaz_inc_ip"))
global constant tDISP  = add_instruction (routine_id ("topaz_display"))
global constant tNOP   = add_instruction (routine_id ("topaz_nop"))
global constant tEND   = add_instruction (routine_id ("topaz_end"))

 -- new instructions to version 2.0
global constant tGINDX = add_instruction (routine_id ("get_index"))
global constant tSINDX = add_instruction (routine_id ("set_index"))
global constant tGSLIC = add_instruction (routine_id ("get_slice"))
global constant tSSLIC = add_instruction (routine_id ("set_slice"))

 -- new instruction to version 2.1
global constant tEQUAL = add_instruction (routine_id ("topaz_equal"))

 -- establish other useful global constants
global constant TOPAZ_VERSION    = 2.1        -- version number
global constant INC_FROM_STACK   = 0          -- for use with tINCIP
global constant NULL_VADDR       = 0          -- for use with tPOP
global constant MAX_PRIORITY     = MAX_EU_INT -- largest priority
global constant DEFAULT_PRIORITY = 1          -- standard priority
